#!/bin/bash

#set -e

LD=ld
DKMS_BUILD_DIR="$( cd -- "$(dirname "$0")" > /dev/null 2>&1 ; pwd -P)"
PROJECT_DIR=$DKMS_BUILD_DIR/../..

#如果后续 KERNEL_HEADER_WHITELIST 白名单频繁修改，考虑单独维护一个.txt文件
KERNEL_HEADER_WHITELIST="types.h kern_levels.h list.h errno.h export.h"

INNOGPU_PURE_HEADERS="hal.h hal_interface.h hal_dma_errors.h innogpu_ion.h kernel_compat.h"
INNOSRVKM_PURE_HEADERS=""
INNODMA_PURE_HEADERS="axi_dmac.h pcie_dmac.h utils.h innodma_debug.h inno_axi_dma_drv.h inno_pcie_dma_drv.h innodma.h"
INNODPU_PURE_HEADERS=""
INNOAUDIO_PURE_HEADERS=""
INNOVPU_PURE_HEADERS=""

INNOGPU_REGIF_HEADERS="g1p_soc_regif.c g0m_soc_regif.c hal_bmc_regif.c"

FIND_FILTER=("-not" "-path" "*innodma/test/*" \
	"-not" "-path" "*innogpu/ion/*" \
	"-not" "-path" "*_pal/*" \
	"-not" "-path" "*_ne/*")

OPEN_SRCS_SUFFIX=_open_srcs.txt
BIN_EXCLUDE_OBJ_SUFFIX=_bin_exclude_objs.tmp

# 模块是否有闭源的二进制
function module_has_private_objs() {
	local module=$1
	bin_exclude_objs=$DKMS_BUILD_DIR/${module}${BIN_EXCLUDE_OBJ_SUFFIX}

	if [ "$module" == "innodpu" ]; then
		objs_dir=$PROJECT_DIR/innosrvkm/display/
		exclude_dir="*innoaudio/*"
	elif [ "$module" == "innoaudio" ]; then
		objs_dir=$PROJECT_DIR/innosrvkm/display/innoaudio
	else
		objs_dir=$PROJECT_DIR/${module}/
		if [ "$module" == "innosrvkm" ]; then
			exclude_dir="*display/*"
		fi
	fi
	find $objs_dir -name "*.o" -not -path "${exclude_dir}" | grep -vFf $bin_exclude_objs | wc -l
}

# 生成链接二进制需要排除的.o文件
# 该函数是通用实现,不要轻易修改,如果模块有特殊需求自行实现
function prepare_bin_exclude_objs() {
	local module=$1
	open_srcs=$DKMS_BUILD_DIR/${module}${OPEN_SRCS_SUFFIX}
	bin_exclude_objs=$DKMS_BUILD_DIR/${module}${BIN_EXCLUDE_OBJ_SUFFIX}

	cat $open_srcs | grep "\.c" > $bin_exclude_objs
	sed -i 's/\.c/.o/' $bin_exclude_objs
}

function get_all_module_pure_header() {
	innoall_pure_headers=$DKMS_BUILD_DIR/innoall_pure_headers.txt

	if [ -f $innoall_pure_headers ]; then
		rm $innoall_pure_headers
	fi

	#todo::: 目前innoaudio的还没有进行整改，也并没有扫描innosrvkm/display/innoaudio/osfunc/目录
	for module in "innogpu" "innosrvkm" "innodma" "innopmbus" "innopower" "innodpu" "innoaudio" "innovpu" "innosr" "innoion";
	do
		inno_module_pure_headers=$DKMS_BUILD_DIR/${module}_pure_headers.txt
		module_pure_header_list=${module^^}_PURE_HEADERS
		IFS=' ' read -ra headers_array <<< "${!module_pure_header_list}"

		if [ -f $inno_module_pure_headers ]; then
			rm $inno_module_pure_headers
		fi

		for pure_header in "${headers_array[@]}"; do
			find $PROJECT_DIR/$module -name $pure_header >> $inno_module_pure_headers
		done

		if [ "$module" == "innogpu" ] || [ "$module" == "innosrvkm" ]; then
			if [ "$module" == "innosrvkm" ]; then
				inno_module_pure_headers_path=$PROJECT_DIR/innosrvkm/services/server/env/linux/
			else
				inno_module_pure_headers_path=$PROJECT_DIR/innogpu/osfunc/
			fi
			find $inno_module_pure_headers_path -name "inno_*.h" >> $inno_module_pure_headers
		fi

		if [ -f $inno_module_pure_headers ]; then
			cat $inno_module_pure_headers >> $innoall_pure_headers
		fi
	done
}

function get_all_module_close_header() {
	innoall_close_headers=$DKMS_BUILD_DIR/innoall_close_headers.txt
	inno_module_headers=$DKMS_BUILD_DIR/inno_module_headers.txt
	inno_open_headers=$DKMS_BUILD_DIR/inno_open_headers.txt

	if [ -f $innoall_close_headers ]; then
		rm $innoall_close_headers
	fi

	#TODO:::not check innosrvkm currently
	for module in "innogpu" "innodma" "innopmbus" "innopower" "innodpu" "innoaudio" "innovpu" "innosr" "innoion";
	do
		open_srcs=$DKMS_BUILD_DIR/${module}${OPEN_SRCS_SUFFIX}
		inno_module_close_headers=$DKMS_BUILD_DIR/${module}_close_headers.txt

		if [ -f $inno_open_headers ]; then
			rm $inno_open_headers
		fi

		grep '\.h' $open_srcs > $inno_open_headers
		for open_heaer in $(grep '\.h' $open_srcs)
		do
			find $PROJECT_DIR/ -name $open_heaer >> $inno_open_headers
		done

		if [ "$module" == "innodpu" ]; then
			find $PROJECT_DIR/innosrvkm/display/ -name "*.h" "${FIND_FILTER[@]}" -not -path "*innoaudio/*" > $inno_module_headers
		elif [ "$module" == "innoaudio" ]; then
			find $PROJECT_DIR/innosrvkm/display/innoaudio/ -name "*.h" "${FIND_FILTER[@]}" > $inno_module_headers
		else
			find $PROJECT_DIR/${module}/ -name "*.h" "${FIND_FILTER[@]}" > $inno_module_headers
			if [ "$module" == "innogpu" ]; then
				IFS=' ' read -ra headers_array <<< $INNOGPU_REGIF_HEADERS
				for regif_header in "${headers_array[@]}"; do
					find $PROJECT_DIR/$module -name $regif_header >> $inno_module_headers
				done
			fi
		fi
		grep -vFf $inno_open_headers $inno_module_headers > $inno_module_close_headers
		cat $inno_module_close_headers >> $innoall_close_headers
	done
}

function check_inc_kernel_header() {
	local check_file=$1
	local check_type=$2

	grep -q "^\s*#include\s*<.*>" $check_file
	if [ $? -eq 0 ]; then
		for header in $(cat "${check_file}" | grep "^\s*#include\s*<.*>" | sed 's/#include\s*<.*\/\([^>]*\)>.*/\1/'); do
			echo $KERNEL_HEADER_WHITELIST | grep -q ${header}
			if [ $? -ne 0 ]; then
				find /usr/src/linux-headers-$(uname -r)/include -name ${header} | grep -q ${header}
				if [ $? -eq 0 ]; then
					if [ "$check_type" == "pure_header" ]; then
						echo "open pure header error: find kernel header $header at $check_file"
					else
						echo "close file error: find kernel header $header at $check_file"
					fi
					exit 1
				fi
			fi
		done
	fi
}

function check_inc_not_pure_header() {
	local check_file=$1

	grep -q "^\s*#include\s*\".*\"" $check_file
	if [ $? -eq 0 ]; then
		for header in $(cat "${check_file}" | grep "^\s*#include\s*\".*\"" | sed 's/^#include "\(.*\)"/\1/'); do
			grep -q ${header} $DKMS_BUILD_DIR/innoall_pure_headers.txt
			if [ $? -ne 0 ]; then
				echo "open pure header file error: find not pure header $header at $check_file"
				exit 1
			fi
		done
	fi
}

function check_inc_not_pure_close_header() {
	local check_file=$1
	grep -q "^\s*#include\s*\".*\"" $check_file
	if [ $? -eq 0 ]; then
		for header in $(cat "${check_file}" | grep "^\s*#include\s*\".*\"" | sed 's/^#include "\(.*\)"/\1/'); do
			grep -q ${header} $DKMS_BUILD_DIR/innoall_pure_headers.txt $DKMS_BUILD_DIR/innoall_close_headers.txt
			if [ $? -ne 0 ]; then
				echo "close file error: find not pure or close header $header at $check_file"
				exit 1
			fi
		done
	fi
}

function check_use_export_symbol() {
	local check_file=$1

	grep -q -E 'EXPORT_SYMBOL|INNO_EXT_SYM|INNO_EXT_TRACEPOINT_SYM' $check_file
	if [ $? -eq 0 ]; then
		echo "close file error: use EXPORT_SYMBOL or INNO_EXT_SYM or INNO_EXT_TRACEPOINT_SYM at $check_file"
		exit 1
	fi
}

# 遍历闭源模块.c文件，检索闭源.c文件中是否引用内核头文件
function scan_module() {
	local module=$1
	objs_name=$DKMS_BUILD_DIR/${module}_objs.txt
	bin_exclude_objs=$DKMS_BUILD_DIR/${module}${BIN_EXCLUDE_OBJ_SUFFIX}

	#inno_module_pure_headers(innogpu/osfunc + INNOXXX_PURE_HEADERS) shall not include kernel header
	#todo::: 目前innoaudio的还没有进行整改，也并没有扫描innosrvkm/display/innoaudio/osfunc/目录
	inno_module_pure_headers=$DKMS_BUILD_DIR/${module}_pure_headers.txt
	if [ -f $inno_module_pure_headers ]; then
		for inno_header in $(cat $inno_module_pure_headers)
		do
			#检查是否包含内核头文件
			check_inc_kernel_header $inno_header "pure_header"
			#检查inno pure header是否包含非inno pure header(包含闭源头文件和其他开源非inno pure header)，
			check_inc_not_pure_header $inno_header
		done
	fi

	inno_module_close_headers=$DKMS_BUILD_DIR/${module}_close_headers.txt
	# todo::: not check innosrvkm currently
	if [ "$module" != "innosrvkm" ]; then
		if [ -f $inno_module_close_headers ]; then
			for inno_header in $(cat $inno_module_close_headers)
			do
				#检查是否包含内核头文件
				check_inc_kernel_header $inno_header
				#检查inno pure header是否包含非inno pure header(包含闭源头文件和其他开源非inno pure header)，
				check_inc_not_pure_close_header $inno_header
			done
		fi
	fi

	cnt=$(module_has_private_objs $module)
	if [[ $cnt != 0 ]]; then
		if [ "$module" == "innodpu" ]; then
			objs_dir=$PROJECT_DIR/innosrvkm/display/
			exclude_dir="*innoaudio/*"
		elif [ "$module" == "innoaudio" ]; then
			objs_dir=$PROJECT_DIR/innosrvkm/display/innoaudio
		else
			objs_dir=$PROJECT_DIR/${module}/
			if [ "$module" == "innosrvkm" ]; then
				exclude_dir="*display/*"
			elif [ "$module" == "innogpu" ]; then
				exclude_dir="*innogpu/ion/*"
			fi
		fi

		find $objs_dir -name "*.o" -not -path "${exclude_dir}" -not -name "*_pal*" | grep -vFf $bin_exclude_objs > $objs_name
		sed -i 's/\.o$/.c/g' $objs_name
		for close_file in $(cat $objs_name)
		do
			#close files shall not include kernel headers
			check_inc_kernel_header $close_file;
			#close files shall only include inno pure headers
			if [ "$module" != "innosrvkm" ]; then # todo::: not check innosrvkm currently
				check_inc_not_pure_close_header $close_file;
			fi
			#extra check close files shall not use EXPORT_SYMBOL, since export.h is at whitelist
			check_use_export_symbol $close_file
		done
	fi
}

function prepare_innogpu_exclude_objs() {
	open_srcs=$DKMS_BUILD_DIR/innogpu${OPEN_SRCS_SUFFIX}
	bin_exclude_objs=$DKMS_BUILD_DIR/innogpu${BIN_EXCLUDE_OBJ_SUFFIX}

	if [[ $CONFIG_INNO_ION != "y" ]]; then
		sed '/ionn*/d' $open_srcs | grep "\.c" > $bin_exclude_objs
	else
		cat $open_srcs | grep "\.c" > $bin_exclude_objs
	fi

	cat $open_srcs | grep "\.S" >> $bin_exclude_objs

	sed -i 's/\.c/.o/' $bin_exclude_objs
	sed -i 's/\.S/.o/' $bin_exclude_objs
}

function prepare_innodma_exclude_objs() {
	prepare_bin_exclude_objs innodma
}

function prepare_innodpu_exclude_objs() {
		prepare_bin_exclude_objs innodpu
}

function prepare_innoaudio_exclude_objs() {
		prepare_bin_exclude_objs innoaudio
}

function prepare_innoion_exclude_objs() {
	prepare_bin_exclude_objs innoion
}

function prepare_innopmbus_exclude_objs() {
	prepare_bin_exclude_objs innopmbus
}

function prepare_innovpu_exclude_objs() {
	prepare_bin_exclude_objs innovpu
}

function prepare_innosr_exclude_objs() {
	prepare_bin_exclude_objs innosr
}

function prepare_innopower_exclude_objs() {
	prepare_bin_exclude_objs innopower
}

function prepare_innoapu_exclude_objs() {
	prepare_bin_exclude_objs innoapu_img_mem
	prepare_bin_exclude_objs innoapu_vha
}

function prepare_innosrvkm_exclude_objs() {
	open_srcs=$DKMS_BUILD_DIR/innosrvkm_open_srcs.txt
	bin_objs=$DKMS_BUILD_DIR/innosrvkm${BIN_EXCLUDE_OBJ_SUFFIX}
	_FILE=${open_srcs}.tmp
	cp $open_srcs $_FILE

	if [[ $PVR_BUILD_DIR = android_container ]]; then
		sed -i '/pvr_sync_ioctl_drm.c/d' $_FILE
	else
		sed -i '/pvr_sync_ioctl_dev.*/d' $_FILE
		sed -i '/ion_sys.*/d' $_FILE
	fi

	if [[ $CONFIG_DRM_INNO_POWER != "y" ]]; then
		sed -i '/pvr_dvfs_device.*/d' $_FILE
	fi

	cat $_FILE | grep "\.c" > $bin_objs

	if [[ $CONFIG_DRM_INNO_DPU = "y" ]]; then
		echo "innogpu_drm.c" >> $bin_objs
	else
		echo "innogpu_drm_without_dpu.c" >> $bin_objs
	fi

	sed -i 's/\.c/.o/' $bin_objs
}

function scan_module_innogpu() {
	scan_module innogpu
}

function scan_module_innodma() {
	scan_module innodma
}

function scan_module_innodpu() {
	scan_module innodpu
}

function scan_module_innoaudio() {
	scan_module innoaudio
}

function scan_module_innoion() {
	scan_module innoion
}

function scan_module_innopmbus() {
	scan_module innopmbus
}

function scan_module_innosrvkm() {
	scan_module innosrvkm
}

function scan_module_innovpu() {
	scan_module innovpu
}

function scan_module_innosr() {
	scan_module innosr
}

function scan_module_innopower() {
	scan_module innopower
}

function make_scan_module() {
	local module=$1
	prepare_${module}_exclude_objs
	echo -e "\nscan module $module:"
	scan_module_${module}
}

function make_scan_all_modules() {

	echo -e "\nstart check dkms code rules ..."
	get_all_module_pure_header
	get_all_module_close_header

	make_scan_module innogpu
	make_scan_module innosrvkm
	make_scan_module innodma

#	if [[ $CONFIG_DRM_INNO_DPU = "y" ]]; then
#		make_scan_module innodpu
#	fi
#
#	if [[ $CONFIG_DRM_INNO_AUDIO = "y" ]]; then
#		make_scan_module innoaudio
#	fi
#
#	if [[ $CONFIG_DRM_INNO_VPU = "y" ]]; then
#		make_scan_module innovpu
#	fi
#
#	if [[ $CONFIG_DRM_INNO_SR = "y" ]]; then
#		make_scan_module innosr
#	fi
#
#	if [[ $CONFIG_INNO_ION = "y" ]]; then
#		make_scan_module innoion
#	fi
#
#	if [[ $CONFIG_DRM_INNO_PMBUS = "y" ]]; then
#		make_scan_module innopmbus
#	fi
#
#	if [[ $CONFIG_DRM_INNO_POWER = "y" ]]; then
#		make_scan_module innopower
#	fi
#
#	if [[ $CONFIG_DRM_INNO_APU = "y" ]]; then
#		make_scan_module innoapu
#	fi

	echo -e "\npass dkms code rules!\n"
}

# 清理
function clean() {
	rm -rf $DKMS_BUILD_DIR/*.tmp
	rm -rf $DKMS_BUILD_DIR/modules_config.sh
	rm -rf $DKMS_BUILD_DIR/*_objs.txt
	rm -rf $DKMS_BUILD_DIR/*_headers.txt
}

make_scan_all_modules
clean
exit 0
